from __future__ import annotations

import warnings
from collections.abc import KeysView, ValuesView
from typing import Any, ClassVar

from coffea.nanoevents.schemas.base import BaseSchema, zip_forms

from atlas_schema.typing_compat import Behavior, Self


class NtupleSchema(BaseSchema):  # type: ignore[misc]
    """Ntuple schema builder

    The Ntuple schema is built from all branches found in the supplied file, based on
    the naming pattern of the branches. The following additional arrays are constructed:

    - n/a
    """

    __dask_capable__ = True

    warn_missing_crossrefs = True
    error_missing_event_ids = False

    event_ids_data: ClassVar[set[str]] = {
        "lumiBlock",
        "averageInteractionsPerCrossing",
        "actualInteractionsPerCrossing",
        "dataTakingYear",
    }
    event_ids_mc: ClassVar[set[str]] = {
        "mcChannelNumber",
        "runNumber",
        "eventNumber",
        "mcEventWeights",
    }
    event_ids: ClassVar[set[str]] = {*event_ids_data, *event_ids_mc}

    mixins: ClassVar[dict[str, str]] = {
        "el": "Electron",
        "jet": "Jet",
        "met": "MissingET",
        "mu": "Muon",
        "pass": "Pass",
        "ph": "Photon",
        "trigPassed": "Trigger",
        "weight": "Weight",
    }

    # These are stored as length-1 vectors unnecessarily
    singletons: ClassVar[set[str]] = set()

    docstrings: ClassVar[dict[str, str]] = {
        "charge": "charge",
        "eta": "pseudorapidity",
        "met": "missing transverse energy [MeV]",
        "mass": "invariant mass [MeV]",
        "pt": "transverse momentum [MeV]",
        "phi": "azimuthal angle",
    }

    def __init__(self, base_form: dict[str, Any], version: str = "latest"):
        super().__init__(base_form)
        self._version = version
        if version == "latest":
            pass
        else:
            pass
        self._form["fields"], self._form["contents"] = self._build_collections(
            self._form["fields"], self._form["contents"]
        )
        self._form["parameters"]["metadata"]["version"] = self._version

    @classmethod
    def v1(cls, base_form: dict[str, Any]) -> Self:
        """Build the NtupleEvents

        For example, one can use ``NanoEventsFactory.from_root("file.root", schemaclass=NtupleSchema.v1)``
        to ensure NanoAODv7 compatibility.
        """
        return cls(base_form, version="1")

    def _build_collections(
        self, field_names: list[str], input_contents: list[Any]
    ) -> tuple[KeysView[str], ValuesView[dict[str, Any]]]:
        branch_forms = dict(zip(field_names, input_contents))

        # parse into high-level records (collections, list collections, and singletons)
        collections = {k.split("_")[0] for k in branch_forms}
        collections -= self.event_ids
        collections -= set(self.singletons)

        # rename needed because easyjet breaks the AMG assumptions
        # https://gitlab.cern.ch/easyjet/easyjet/-/issues/246
        for k in list(branch_forms):
            if "NOSYS" not in k:
                continue
            branch_forms[k.replace("_NOSYS", "") + "_NOSYS"] = branch_forms.pop(k)

        # these are collections with systematic variations
        subcollections = {
            k.split("__")[0].split("_", 1)[1].replace("_NOSYS", "")
            for k in branch_forms
            if "NOSYS" in k
        }

        # Check the presence of the event_ids
        missing_event_ids = [
            event_id for event_id in self.event_ids if event_id not in branch_forms
        ]

        if len(missing_event_ids) > 0:
            if self.error_missing_event_ids:
                msg = f"There are missing event ID fields: {missing_event_ids} \n\n\
                    The event ID fields {self.event_ids} are necessary to perform sub-run identification \
                    (e.g. for corrections and sub-dividing data during different detector conditions),\
                    to cross-validate MC and Data (i.e. matching events for comparison), and to generate event displays. \
                    It's advised to never drop these branches from the dataformat.\n\n\
                    This error can be demoted to a warning by setting the class level variable error_missing_event_ids to False."
                raise RuntimeError(msg)
            warnings.warn(
                f"Missing event_ids : {missing_event_ids}",
                RuntimeWarning,
                stacklevel=2,
            )

        output = {}

        # first, register singletons (event-level, others)
        for name in {*self.event_ids, *self.singletons}:
            if name in missing_event_ids:
                continue
            output[name] = branch_forms[name]

        # next, go through and start grouping up collections
        for name in collections:
            mixin = self.mixins.get(name, "NanoCollection")
            content = {}
            used = set()

            for subname in subcollections:
                prefix = f"{name}_{subname}_"
                used.update({k for k in branch_forms if k.startswith(prefix)})
                subcontent = {
                    k[len(prefix) :]: branch_forms[k]
                    for k in branch_forms
                    if k.startswith(prefix)
                }
                if subcontent:
                    # create the nominal version
                    content[subname] = branch_forms[f"{prefix}NOSYS"]
                    # create a collection of the systematic variations for the given variable
                    content[f"{subname}_syst"] = zip_forms(
                        subcontent, f"{name}_syst", record_name="NanoCollection"
                    )

            content.update(
                {
                    k[len(name) + 1 :]: branch_forms[k]
                    for k in branch_forms
                    if k.startswith(name + "_") and k not in used
                }
            )

            if not used and not content:
                warnings.warn(
                    f"I identified a branch that likely does not have any leaves: '{name}'. I will treat this as a 'singleton'. To suppress this warning next time, please define your singletons explicitly.",
                    RuntimeWarning,
                    stacklevel=2,
                )
                self.singletons.add(name)
                output[name] = branch_forms[name]

            else:
                output[name] = zip_forms(content, name, record_name=mixin)

            output[name].setdefault("parameters", {})
            output[name]["parameters"].update({"collection_name": name})

            if output[name]["class"] == "ListOffsetArray":
                parameters = output[name]["content"]["fields"]
                contents = output[name]["content"]["contents"]
            elif output[name]["class"] == "RecordArray":
                parameters = output[name]["fields"]
                contents = output[name]["contents"]
            elif output[name]["class"] == "NumpyArray":
                # these are singletons that we just pass through
                continue
            else:
                msg = f"Unhandled class {output[name]['class']}"
                raise RuntimeError(msg)
            # update docstrings as needed
            # NB: must be before flattening for easier logic
            for index, parameter in enumerate(parameters):
                if "parameters" not in contents[index]:
                    continue

                parsed_name = parameter.replace("_NOSYS", "")
                contents[index]["parameters"]["__doc__"] = self.docstrings.get(
                    parsed_name,
                    contents[index]["parameters"].get(
                        "__doc__", "no docstring available"
                    ),
                )

        return output.keys(), output.values()

    @classmethod
    def behavior(cls) -> Behavior:
        """Behaviors necessary to implement this schema"""
        from atlas_schema.methods import behavior as roaster

        return roaster



"""Mixins for the Ntuple schema"""

from __future__ import annotations

from functools import reduce
from operator import ior

import awkward
import particle
from coffea.nanoevents.methods import base, candidate, vector
from dask_awkward import dask_method

from atlas_schema.enums import PhotonID
from atlas_schema.typing_compat import Behavior

behavior: Behavior = {}
behavior.update(base.behavior)
# vector behavior is included in candidate behavior
behavior.update(candidate.behavior)


class NtupleEvents(behavior["NanoEvents"]):  # type: ignore[misc, valid-type, name-defined]
    def __repr__(self):
        return f"<event {getattr(self, 'runNumber', '??')}:\
                {getattr(self, 'eventNumber', '??')}:\
                {getattr(self, 'mcChannelNumber', '??')}>"


behavior["NanoEvents"] = NtupleEvents


def _set_repr_name(classname):
    def namefcn(_self):
        return classname

    behavior[("__typestr__", classname)] = classname[0].lower() + classname[1:]
    behavior[classname].__repr__ = namefcn


@awkward.mixin_class(behavior)
class Weight(base.NanoCollection, base.Systematic): ...


_set_repr_name("Weight")


@awkward.mixin_class(behavior)
class Pass(base.NanoCollection, base.Systematic): ...


_set_repr_name("Pass")

behavior.update(
    awkward._util.copy_behaviors("PtEtaPhiMLorentzVector", "Particle", behavior)
)


@awkward.mixin_class(behavior)
class Particle(vector.PtEtaPhiMLorentzVector):
    """Generic particle collection that has Lorentz vector properties

    Also handles the following additional branches:
    - '{obj}_select'
    """

    @property
    def mass(self):
        r"""Invariant mass (+, -, -, -)

        :math:`\sqrt{t^2-x^2-y^2-z^2}`
        """
        return self["mass"] / 1.0e3

    @dask_method
    def passes(self, name):
        return self[f"select_{name}"] == 1

    @passes.dask
    def passes(self, dask_array, name):
        return dask_array[f"select_{name}"] == 1

    # NB: fields with the name 'pt' take precedence over this
    # @dask_property
    # def pt(self):
    #     print('inside non-dask prop')
    #     return self["pt_NOSYS"]

    # @pt.dask
    # def pt(self, dask_array):
    #     branch = 'pt'
    #     print('inside dask prop')
    #     variation = dask_array._events().metadata.get("systematic", "NOSYS")
    #     with contextlib.suppress(Exception):
    #         return dask_array[f"{branch}_{variation}"]

    #     if variation != "NOSYS":
    #         with contextlib.suppress(Exception):
    #             return dask_array[f"{branch}_NOSYS"]

    #     return dask_array[branch]


_set_repr_name("Particle")

ParticleArray.ProjectionClass2D = vector.TwoVectorArray  # noqa: F821
ParticleArray.ProjectionClass3D = vector.ThreeVectorArray  # noqa: F821
ParticleArray.ProjectionClass4D = ParticleArray  # noqa: F821
ParticleArray.MomentumClass = vector.LorentzVectorArray  # noqa: F821


behavior.update(awkward._util.copy_behaviors("PolarTwoVector", "MissingET", behavior))


@awkward.mixin_class(behavior)
class MissingET(vector.PolarTwoVector, base.NanoCollection, base.Systematic):
    @property
    def r(self):
        """Distance from origin in XY plane"""
        return self["met"]


_set_repr_name("MissingET")

MissingETArray.ProjectionClass2D = MissingETArray  # noqa: F821
MissingETArray.ProjectionClass3D = vector.SphericalThreeVectorArray  # noqa: F821
MissingETArray.ProjectionClass4D = vector.LorentzVectorArray  # noqa: F821
MissingETArray.MomentumClass = MissingETArray  # noqa: F821

behavior.update(awkward._util.copy_behaviors("Particle", "Photon", behavior))


@awkward.mixin_class(behavior)
class Photon(Particle, base.NanoCollection, base.Systematic):
    @property
    def mass(self):
        """Return zero mass for photon."""
        return awkward.zeros_like(self.pt)

    @property
    def charge(self):
        """Return zero charge for photon."""
        return awkward.zeros_like(self.pt)

    @property
    def isEM(self):
        return self.isEM_syst.NOSYS == 0

    def pass_isEM(self, words: list[PhotonID]):
        # 0 is pass, 1 is fail
        return (
            self.isEM_syst.NOSYS & reduce(ior, (1 << word.value for word in words))
        ) == 0


_set_repr_name("Photon")

PhotonArray.ProjectionClass2D = vector.TwoVectorArray  # noqa: F821
PhotonArray.ProjectionClass3D = vector.ThreeVectorArray  # noqa: F821
PhotonArray.ProjectionClass4D = PhotonArray  # noqa: F821
PhotonArray.MomentumClass = vector.LorentzVectorArray  # noqa: F821

behavior.update(awkward._util.copy_behaviors("Particle", "Electron", behavior))


@awkward.mixin_class(behavior)
class Electron(Particle, base.NanoCollection, base.Systematic):
    @property
    def mass(self):
        """Electron mass in GeV"""
        return particle.literals.e_minus.mass / 1.0e3


_set_repr_name("Electron")

ElectronArray.ProjectionClass2D = vector.TwoVectorArray  # noqa: F821
ElectronArray.ProjectionClass3D = vector.ThreeVectorArray  # noqa: F821
ElectronArray.ProjectionClass4D = ElectronArray  # noqa: F821
ElectronArray.MomentumClass = vector.LorentzVectorArray  # noqa: F821

behavior.update(awkward._util.copy_behaviors("Particle", "Muon", behavior))


@awkward.mixin_class(behavior)
class Muon(Particle, base.NanoCollection, base.Systematic):
    @property
    def mass(self):
        """Muon mass in GeV"""
        return particle.literals.mu_minus.mass / 1.0e3


_set_repr_name("Muon")

MuonArray.ProjectionClass2D = vector.TwoVectorArray  # noqa: F821
MuonArray.ProjectionClass3D = vector.ThreeVectorArray  # noqa: F821
MuonArray.ProjectionClass4D = MuonArray  # noqa: F821
MuonArray.MomentumClass = vector.LorentzVectorArray  # noqa: F821

behavior.update(awkward._util.copy_behaviors("Particle", "Tau", behavior))


@awkward.mixin_class(behavior)
class Tau(Particle, base.NanoCollection, base.Systematic):
    @property
    def mass(self):
        """Tau mass in GeV"""
        return particle.literals.tau_minus.mass / 1.0e3


_set_repr_name("Tau")

TauArray.ProjectionClass2D = vector.TwoVectorArray  # noqa: F821
TauArray.ProjectionClass3D = vector.ThreeVectorArray  # noqa: F821
TauArray.ProjectionClass4D = TauArray  # noqa: F821
TauArray.MomentumClass = vector.LorentzVectorArray  # noqa: F821


behavior.update(awkward._util.copy_behaviors("Particle", "Jet", behavior))


@awkward.mixin_class(behavior)
class Jet(Particle, base.NanoCollection, base.Systematic): ...


_set_repr_name("Jet")

JetArray.ProjectionClass2D = vector.TwoVectorArray  # noqa: F821
JetArray.ProjectionClass3D = vector.ThreeVectorArray  # noqa: F821
JetArray.ProjectionClass4D = JetArray  # noqa: F821
JetArray.MomentumClass = vector.LorentzVectorArray  # noqa: F821

__all__ = [
    "Electron",
    "Jet",
    "MissingET",
    "Muon",
    "NtupleEvents",
    "Particle",
    "Pass",
    "Photon",
    "Weight",
]



from __future__ import annotations

from enum import Enum
from typing import TypeVar, Union, cast

import awkward as ak
import dask_awkward as dak

Array = TypeVar("Array", bound=Union[dak.Array, ak.Array])
_E = TypeVar("_E", bound=Enum)


def isin(element: Array, test_elements: dak.Array | ak.Array, axis: int = -1) -> Array:
    """
    Find test_elements in element. Similar in API as :func:`numpy.isin`.

    Calculates `element in test_elements`, broadcasting over *element elements only*. Returns a boolean array of the same shape as *element* that is `True` where an element of *element* is in *test_elements* and `False` otherwise.

    This works by first transforming *test_elements* to an array with one more
    dimension than the *element*, placing the *test_elements* at *axis*, and then doing a
    comparison.

    Args:
        element (dak.Array or ak.Array): input array of values.
        test_elements (dak.Array or ak.Array): one-dimensional set of values against which to test each value of *element*.
        axis (int): the axis along which the comparison is performed

    Returns:
        dak.Array or ak.Array: result of comparison for test_elements in *element*

    Example:
        >>> import awkward as ak
        >>> import atlas_schema as ats
        >>> truth_origins = ak.Array([[1, 2, 3], [4], [5, 6, 7], [1]])
        >>> prompt_origins = ak.Array([1, 2, 7])
        >>> ats.isin(truth_origins, prompt_origins).to_list()
        [[True, True, False], [False], [False, False, True], [True]]
    """
    assert test_elements.ndim == 1, "test_elements must be one-dimensional"
    assert axis >= -1, "axis must be -1 or positive-valued"
    assert axis < element.ndim + 1, "axis too large for the element"

    # First, build up the transformation, with slice(None) indicating where to stick the test_elements
    reshaper: list[None | slice] = [None] * element.ndim
    axis = element.ndim if axis == -1 else axis
    reshaper.insert(axis, slice(None))

    # Note: reshaper needs to be a tuple for indexing purposes
    return cast(Array, ak.any(element == test_elements[tuple(reshaper)], axis=-1))



"""
Copyright (c) 2024 Giordon Stark. All rights reserved.

atlas_schema: Collection of utilities and helper functions for HEP ATLAS analysers
"""

from __future__ import annotations

import warnings

from atlas_schema._version import version as __version__
from atlas_schema.enums import ParticleOrigin, PhotonID
from atlas_schema.utils import isin

warnings.filterwarnings("ignore", module="coffea.*")

__all__ = ["ParticleOrigin", "PhotonID", "__version__", "isin"]
